use super::{api, OIDCError, OIDCErrorType, Request};
use crate::{
    realm::RealmHelper,
    schema,
    server::session::SessionHelper,
    user::{UserError, UserExt},
    UIDCError,
};
use microrm::prelude::*;

fn do_authorization_code<'l>(
    realm: &microrm::Stored<schema::Realm>,
    rhelper: &RealmHelper,
    client: &microrm::Stored<schema::Client>,
    treq: &api::TokenRequestBody,
) -> Result<tide::Response, OIDCError<'l>> {
    let Some(code) = treq.code.as_ref() else {
        return Err(OIDCError(
            OIDCErrorType::InvalidRequest,
            "no authorization code provided".into(),
            None,
        ));
    };
    let code = realm
        .auth_codes
        .keyed((realm.id(), client.id(), code))
        .get()?
        .ok_or(OIDCError(
            OIDCErrorType::AccessDenied,
            "invalid authorization code".into(),
            None,
        ))?;

    let now = std::time::SystemTime::now();
    if code.expiry < now {
        return Err(OIDCError(
            OIDCErrorType::AccessDenied,
            "expired authorization code".into(),
            None,
        ));
    }

    let user = realm
        .users
        .with(schema::UserID::default(), code.user)
        .first()
        .get()?
        .ok_or(OIDCError(
            OIDCErrorType::ServerError,
            "could not find user".into(),
            None,
        ))?;

    let access_token = rhelper
        .generate_access_token(
            client,
            &user,
            code.scopes.as_ref().iter().map(String::as_str),
        )
        .map_err(|e| {
            OIDCError(
                OIDCErrorType::ServerError,
                format!("error signing key: {e}").into(),
                None,
            )
        })?;

    Ok(tide::Response::builder(200)
        .content_type(tide::http::mime::JSON)
        .body(
            serde_json::to_value(api::TokenResponse {
                access_token: access_token.as_str(),
                token_type: "bearer",
                refresh_token: None,
                scope: None,
            })
            .unwrap(),
        )
        .build())
}

fn do_refresh_token<'l>(
    _realm: &microrm::Stored<schema::Realm>,
    rhelper: &RealmHelper,
    client: &microrm::Stored<schema::Client>,
    treq: &api::TokenRequestBody,
) -> Result<tide::Response, OIDCError<'l>> {
    let Some(rtoken) = treq.refresh_token.as_ref() else {
        return Err(OIDCError(
            OIDCErrorType::InvalidRequest,
            "no refresh_token given".into(),
            None,
        ));
    };

    let (access, refresh) = match rhelper.trade_refresh_token(client, rtoken.as_str()) {
        Ok((a, r)) => (a, r),
        Err(e) => {
            return Err(OIDCError(
                OIDCErrorType::InvalidRequest,
                format!("could not trade refresh token: {e}").into(),
                None,
            ))
        }
    };
    Ok(tide::Response::builder(200)
        .content_type(tide::http::mime::JSON)
        .body(
            serde_json::to_value(api::TokenResponse {
                access_token: access.as_str(),
                token_type: "bearer",
                refresh_token: Some(refresh.as_str()),
                scope: None,
            })
            .unwrap(),
        )
        .build())
}

fn do_direct_grant<'l>(
    realm: &microrm::Stored<schema::Realm>,
    rhelper: &RealmHelper,
    client: &microrm::Stored<schema::Client>,
    treq: &api::TokenRequestBody,
) -> Result<tide::Response, OIDCError<'l>> {
    // first thing to check: does the client have the direct grant type enabled?
    if !client.direct_grant_enabled {
        return Err(OIDCError(
            OIDCErrorType::UnsupportedResponseType,
            "client does not have direct grants enabled".into(),
            None,
        ));
    }

    let Some(username) = treq.username.as_ref() else {
        return Err(OIDCError(
            OIDCErrorType::InvalidRequest,
            "no username provided".into(),
            None,
        ));
    };
    let Some(password) = treq.password.as_ref() else {
        return Err(OIDCError(
            OIDCErrorType::InvalidRequest,
            "no password provided".into(),
            None,
        ));
    };

    let Some(user) = realm
        .users
        .with(schema::User::Username, username)
        .first()
        .get()?
    else {
        return Err(OIDCError(
            OIDCErrorType::AccessDenied,
            "no such user".into(),
            None,
        ));
    };

    // verify that we don't accidentally log in someone with username/password who has MFA..
    match user
        .auth
        .with(schema::AuthChallenge::Enabled, true)
        .count()?
    {
        0 => {
            return Err(OIDCError(
                OIDCErrorType::AccessDenied,
                "user has no associated password".into(),
                None,
            ))
        }
        1 => (),
        _ => {
            return Err(OIDCError(
                OIDCErrorType::AccessDenied,
                "user has MFA enabled".into(),
                None,
            ))
        }
    }

    match user.verify_challenge_by_type(schema::AuthChallengeType::Password, password.as_bytes()) {
        Err(UIDCError::UserError(UserError::NoSuchChallenge)) => Err(OIDCError(
            OIDCErrorType::AccessDenied,
            "user has no associated password".into(),
            None,
        )),
        Err(UIDCError::DatabaseError(e)) => Err(e.into()),
        Err(e) => Err(OIDCError(
            OIDCErrorType::ServerError,
            format!("internal error: {e}").into(),
            None,
        )),
        Ok(false) => Err(OIDCError(
            OIDCErrorType::AccessDenied,
            "password authentication failed".into(),
            None,
        )),
        Ok(true) => {
            let scopes = treq.scope.as_deref().unwrap_or("").split_whitespace();
            let Ok(access_token) = rhelper.generate_access_token(client, &user, scopes.clone())
            else {
                return Err(OIDCError(
                    OIDCErrorType::ServerError,
                    "could not generate access token".into(),
                    None,
                ));
            };
            let Ok(refresh_token) = rhelper.generate_refresh_token(client, &user, scopes) else {
                return Err(OIDCError(
                    OIDCErrorType::ServerError,
                    "could not generate access token".into(),
                    None,
                ));
            };

            Ok(tide::Response::builder(200)
                .content_type(tide::http::mime::JSON)
                .body(
                    serde_json::to_vec(&api::TokenResponse {
                        access_token: access_token.as_str(),
                        token_type: "bearer",
                        refresh_token: Some(refresh_token.as_str()),
                        scope: None,
                    })
                    .unwrap(),
                )
                .build())
        }
    }
}

pub(super) async fn do_token<'l>(mut request: Request) -> Result<tide::Response, OIDCError<'l>> {
    let shelper = SessionHelper::new(&request);
    let realm = shelper
        .get_realm()
        .map_err(|_| OIDCError(OIDCErrorType::InvalidRequest, "no such realm".into(), None))?;

    let treq: api::TokenRequestBody = request.body_form().await.map_err(|e| {
        OIDCError(
            OIDCErrorType::InvalidRequest,
            format!("could not parse form body: {e}").into(),
            None,
        )
    })?;

    // TODO: support HTTP basic auth for client authentication instead of treq.client_id
    let client_name = treq.client_id.as_ref().ok_or(OIDCError(
        OIDCErrorType::InvalidRequest,
        "no client given".into(),
        None,
    ))?;
    let client = realm
        .clients
        .keyed((realm.id(), client_name))
        .get()?
        .ok_or(OIDCError(
            OIDCErrorType::InvalidRequest,
            format!("unknown client name {client_name}").into(),
            None,
        ))?;

    let rhelper = request.state().core.realms.get_helper(realm.id()).unwrap();

    if treq.grant_type == "authorization_code" {
        do_authorization_code(&realm, rhelper.as_ref(), &client, &treq)
    } else if treq.grant_type == "refresh_token" {
        do_refresh_token(&realm, rhelper.as_ref(), &client, &treq)
    } else if treq.grant_type == "password" {
        do_direct_grant(&realm, rhelper.as_ref(), &client, &treq)
    } else {
        Err(OIDCError(
            OIDCErrorType::InvalidRequest,
            format!("unknown grant_type {}", treq.grant_type).into(),
            None,
        ))
    }
}
